import os
import logging
import threading
from datetime import datetime, timedelta
from google.cloud.storage import Blob
from google.cloud.storage.client import Client
from google.api_core.exceptions import TooManyRequests, InternalServerError
from google.api_core.exceptions import ServiceUnavailable
from google.resumable_media.common import DataCorruption, InvalidResponse

from servicelayer.archive.virtual import VirtualArchive
from servicelayer.archive.util import checksum, ensure_path, ensure_posix_path
from servicelayer.util import service_retries, backoff

log = logging.getLogger(__name__)
FAILURES = (
    TooManyRequests,
    InternalServerError,
    ServiceUnavailable,
    DataCorruption,
    InvalidResponse,
)


class GoogleStorageArchive(VirtualArchive):
    TIMEOUT = 84600

    def __init__(self, bucket=None):
        super(GoogleStorageArchive, self).__init__(bucket)
        log.info("Archive: gs://%s", bucket)
        self._bucket = bucket
        self.local = threading.local()

    @property
    def bucket(self):
        if not hasattr(self.local, "bucket"):
            client = Client()
            self.local.bucket = client.lookup_bucket(self._bucket)
            if self.bucket is None:
                self.local.bucket = client.create_bucket(self._bucket)
                self.upgrade()
        return self.local.bucket

    def upgrade(self):
        policy = {
            "origin": ["*"],
            "method": ["GET", "HEAD", "OPTIONS"],
            "responseHeader": ["*"],
            "maxAgeSeconds": self.TIMEOUT,
        }
        self.bucket.cors = [policy]
        self.bucket.update()

    def _locate_contenthash(self, content_hash):
        """Check if a file with the given hash exists on S3."""
        if content_hash is None:
            return
        prefix = self._get_prefix(content_hash)
        if prefix is None:
            return

        # First, check the standard file name:
        blob = Blob(os.path.join(prefix, "data"), self.bucket)
        if blob.exists():
            return blob

        # Second, iterate over all file names:
        for blob in self.bucket.list_blobs(max_results=1, prefix=prefix):
            return blob

    def _locate_key(self, key):
        if key is None:
            return
        blob = Blob(key, self.bucket)
        if blob.exists():
            return blob

    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Store the file located at the given path on Google, based on a path
        made up from its SHA1 content hash."""
        file_path = ensure_path(file_path)
        if content_hash is None:
            content_hash = checksum(file_path)

        if content_hash is None:
            return

        file_path = ensure_posix_path(file_path)
        for attempt in service_retries():
            try:
                # blob = self._locate_contenthash(content_hash)
                # if blob is not None:
                #     return content_hash

                path = os.path.join(self._get_prefix(content_hash), "data")
                blob = Blob(path, self.bucket)
                blob.upload_from_filename(file_path, content_type=mime_type)
                return content_hash
            except FAILURES:
                log.exception("Store error in GS")
                backoff(failures=attempt)

    def load_file(self, content_hash, file_name=None, temp_path=None):
        """Retrieve a file from Google storage and put it onto the local file
        system for further processing."""
        for attempt in service_retries():
            try:
                blob = self._locate_contenthash(content_hash)
                if blob is not None:
                    path = self._local_path(content_hash, file_name, temp_path)
                    blob.download_to_filename(path)
                    return path
            except FAILURES:
                log.exception("Load error in GS")
                backoff(failures=attempt)

        # Returns None for "persistent error" as well as "file not found" :/
        log.debug("[%s] not found, or the backend is down.", content_hash)

    def delete_file(self, content_hash):
        """Check if a file with the given hash exists on S3."""
        if content_hash is None:
            return
        prefix = self._get_prefix(content_hash)
        if prefix is None:
            return

        # Iterate over all file names:
        for blob in self.bucket.list_blobs(max_results=1, prefix=prefix):
            for attempt in service_retries():
                try:
                    blob.delete()
                except FAILURES:
                    log.exception("Delete error in GS")
                    backoff(failures=attempt)

    def generate_url(self, content_hash, file_name=None, mime_type=None, expire=None):
        blob = self._locate_contenthash(content_hash)
        if blob is None:
            return
        disposition = None
        if file_name is not None:
            disposition = "inline; filename=%s" % file_name
        if expire is None:
            expire = datetime.utcnow() + timedelta(seconds=self.TIMEOUT)
        return blob.generate_signed_url(
            expire, response_type=mime_type, response_disposition=disposition
        )



from setuptools import setup, find_packages


setup(
    name="servicelayer",
    version="1.16.1",
    description="Basic remote service functions for alephdata components",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="storage files s3",
    author="Organized Crime and Corruption Reporting Project",
    author_email="data@occrp.org",
    url="http://github.com/alephdata/servicelayer",
    license="MIT",
    packages=find_packages(exclude=["ez_setup", "examples", "test"]),
    namespace_packages=[],
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        "banal >= 1.0.1",
        "normality >= 2.1.1",
        "fakeredis == 1.4.5",
        "sqlalchemy >= 1.3",
    ],
    extras_require={
        "amazon": ["boto3 >= 1.11.9, <2.0.0"],
        "google": [
            "grpcio >= 1.32.0, <2.0.0",
            "google-cloud-storage >= 1.31.0, <2.0.0",
        ],
        "dev": [
            "twine",
            "moto",
            "boto3 >= 1.11.9, <2.0.0",
            "pytest >= 3.6",
            "coverage",
            "pytest-cov",
        ],
    },
    test_suite="tests",
    entry_points={
        "servicelayer.test": ["test = servicelayer.extensions:get_extensions"]
    },
)



from unittest import TestCase

from moto import mock_s3

from servicelayer.archive import init_archive
from servicelayer.archive.util import checksum, ensure_path


class S3ArchiveTest(TestCase):
    def setUp(self):
        self.mock = mock_s3()
        self.mock.start()
        self.archive = init_archive("s3", bucket="foo")
        self.file = ensure_path(__file__)

    def tearDown(self):
        self.mock.stop()

    def test_basic_archive(self):
        checksum_ = checksum(self.file)
        assert checksum_ is not None, checksum_
        out = self.archive.archive_file(self.file)
        assert checksum_ == out, (checksum_, out)
        out2 = self.archive.archive_file(self.file)
        assert out == out2, (out, out2)

    def test_basic_archive_with_checksum(self):
        checksum_ = "banana"
        out = self.archive.archive_file(self.file, checksum_)
        assert checksum_ == out, (checksum_, out)

    def test_generate_url(self):
        out = self.archive.archive_file(self.file)
        url = self.archive.generate_url(out)
        # assert False, url
        assert url is not None, url

    def test_load_file(self):
        out = self.archive.archive_file(self.file)
        path = self.archive.load_file(out)
        assert path is not None, path
        assert path.is_file(), path

    def test_cleanup_file(self):
        out = self.archive.archive_file(self.file)
        self.archive.cleanup_file(out)
        path = self.archive.load_file(out)
        assert path.is_file(), path
        self.archive.cleanup_file(out)
        assert not path.exists(), path

    def test_delete_file(self):
        out = self.archive.archive_file(self.file)
        path = self.archive.load_file(out)
        assert path is not None, path
        self.archive.cleanup_file(out)
        self.archive.delete_file(out)
        path = self.archive.load_file(out)
        assert path is None, path



from unittest import TestCase

from servicelayer.cache import get_redis
from servicelayer.jobs import Job, Stage, Task
from servicelayer.reporting import Reporter, OP_REPORT, Status
from servicelayer.util import dump_json


class ReportingTest(TestCase):
    def setUp(self):
        self.conn = get_redis()
        self.dataset = "my-dataset"
        self.operation = "OP_FOO"
        self.job = Job.create(self.conn, self.dataset)

    def _get_task(self):
        """return the newly created reporting task"""
        return Stage.get_task(self.conn, [OP_REPORT])

    def test_reporting(self):
        payload = {"msg": "I do report."}

        reporter = Reporter(stage=self.job.get_stage(self.operation))

        # start
        reporter.start(**payload)
        task = self._get_task()

        self.assertEqual(task.stage.stage, OP_REPORT)

        data = task.payload
        values = {
            "msg": "I do report.",
            "status": Status.START,
            "has_error": False,
            "operation": self.operation,
            "dataset": self.dataset,
            "updated_at": None,
            "start_at": None,
            "job": self.job.id,
        }
        for key, value in values.items():
            self.assertIn(key, data)
            if value is not None:
                self.assertEqual(value, data[key])

        # success
        reporter.end(**payload)
        task = self._get_task()
        self.assertEqual(Status.SUCCESS, task.payload["status"])
        self.assertIn("success_at", task.payload)

        # failing
        reporter.error(Exception("An error occured!"), **payload)
        task = self._get_task()
        data = task.payload
        self.assertEqual(Status.ERROR, data["status"])
        self.assertTrue(data["has_error"])
        self.assertIn("error_at", data)
        self.assertEqual("Exception", data["error_name"])
        self.assertEqual("An error occured!", data["error_msg"])

    def test_reporting_from_task(self):
        task_data = {
            "dataset": self.dataset,
            "job": self.job.id,
            "stage": self.job.get_stage(self.operation).stage,
            "payload": {"entities": ["a", "b", "c"]},
        }
        task = Task.unpack(self.conn, dump_json(task_data))
        serialized = task.serialize()
        reporter = Reporter(task=task)
        reporter.start()

        task = self._get_task()
        self.assertEqual(task.stage.stage, OP_REPORT)

        data = task.payload
        values = {
            "status": Status.START,
            "has_error": False,
            "operation": self.operation,
            "dataset": self.dataset,
            "updated_at": None,
            "start_at": None,
            "job": self.job.id,
            "task": serialized,
        }
        for key, value in values.items():
            self.assertIn(key, data)
            if value is not None:
                self.assertEqual(value, data[key])
